package com.burtbeckwith.dynamicdomain

import org.codehaus.groovy.control.CompilerConfiguration
import org.codehaus.groovy.grails.commons.ApplicationHolder as AH
import org.codehaus.groovy.grails.commons.DomainClassArtefactHandler
import org.codehaus.groovy.grails.commons.GrailsDomainClass
import org.codehaus.groovy.grails.compiler.injection.ClassInjector
import org.codehaus.groovy.grails.compiler.injection.DefaultGrailsDomainClassInjector
import org.codehaus.groovy.grails.compiler.injection.GrailsAwareClassLoader
import org.codehaus.groovy.grails.orm.hibernate.ConfigurableLocalSessionFactoryBean
import org.codehaus.groovy.grails.plugins.DomainClassGrailsPlugin
import org.codehaus.groovy.grails.plugins.orm.hibernate.HibernatePluginSupport
import org.codehaus.groovy.grails.validation.GrailsDomainClassValidator

import org.springframework.beans.factory.config.MethodInvokingFactoryBean
import org.springframework.beans.factory.config.RuntimeBeanReference
import org.springframework.beans.factory.support.AbstractBeanDefinition
import org.springframework.beans.factory.support.GenericBeanDefinition
import org.springframework.util.ReflectionUtils

/**
 * Compiles and registers domain classes at runtime.
 *
 * @author <a href='mailto:burt@burtbeckwith.com'>Burt Beckwith</a>
 */
class DynamicDomainService {

	static transactional = false

	void registerNewDomainClass(String code) {
		def application = AH.application

		def clazz = compile(code, application)

		// register it as if it was a class under grails-app/domain
		GrailsDomainClass dc = application.addArtefact(DomainClassArtefactHandler.TYPE, clazz)

		def ctx = application.mainContext

		registerBeans ctx, dc

		wireMetaclass ctx, dc

		updateSessionFactory ctx
	}

	private Class compile(String code, application) {
		Class clazz = new DynamicClassLoader().parseClass(code)
		application.classLoader.setClassCacheEntry clazz // TODO hack
		clazz
	}

	// this is typically done in DomainClassGrailsPlugin.doWithSpring
	private void registerBeans(ctx, GrailsDomainClass dc) {

		ctx.registerBeanDefinition dc.fullName,
			new GenericBeanDefinition(
				beanClass: dc.clazz,
				scope: AbstractBeanDefinition.SCOPE_PROTOTYPE,
				autowireMode:AbstractBeanDefinition.AUTOWIRE_BY_NAME)

		GenericBeanDefinition beanDef = new GenericBeanDefinition(
			beanClass: MethodInvokingFactoryBean,
			lazyInit: true)
		setBeanProperty beanDef, 'targetObject', new RuntimeBeanReference('grailsApplication', true)
		setBeanProperty beanDef, 'targetMethod', 'getArtefact'
		setBeanProperty beanDef, 'arguments', [DomainClassArtefactHandler.TYPE, dc.fullName]
		ctx.registerBeanDefinition "${dc.fullName}DomainClass", beanDef

		beanDef = new GenericBeanDefinition(
			beanClass: MethodInvokingFactoryBean,
			lazyInit: true)
		setBeanProperty beanDef, 'targetObject', new RuntimeBeanReference("${dc.fullName}DomainClass")
		setBeanProperty beanDef, 'targetMethod', 'getClazz'
		ctx.registerBeanDefinition "${dc.fullName}PersistentClass", beanDef

		beanDef = new GenericBeanDefinition(
			beanClass: GrailsDomainClassValidator,
			lazyInit: true)
		setBeanProperty beanDef, 'messageSource', new RuntimeBeanReference('messageSource')
		setBeanProperty beanDef, 'domainClass', new RuntimeBeanReference("${dc.fullName}DomainClass")
		setBeanProperty beanDef, 'grailsApplication', new RuntimeBeanReference('grailsApplication', true)
		ctx.registerBeanDefinition "${dc.fullName}Validator", beanDef
	}

	private void setBeanProperty(GenericBeanDefinition bean, String name, value) {
		bean.propertyValues.addPropertyValue name, value
	}

	private void wireMetaclass(ctx, GrailsDomainClass dc) {
		def fakeApplication = new FakeApplication(dc)
		DomainClassGrailsPlugin.enhanceDomainClasses fakeApplication, ctx
		HibernatePluginSupport.enhanceSessionFactory ctx.sessionFactory, fakeApplication, ctx
	}

	// creates a new session factory so new classes can reference existing, and then replaces
	// the data in the original session factory with the new combined data
	private void updateSessionFactory(ctx) {
		def sessionFactoryBean = ctx.getBean('&sessionFactory')
		def newSessionFactoryFactory = new ConfigurableLocalSessionFactoryBean(
			dataSource: ctx.dataSource,
			configLocations: getFieldValue(sessionFactoryBean, 'configLocations'),
			configClass: getFieldValue(sessionFactoryBean, 'configClass'),
			hibernateProperties: getFieldValue(sessionFactoryBean, 'hibernateProperties'),
			grailsApplication: ctx.grailsApplication,
			lobHandler: getFieldValue(sessionFactoryBean, 'lobHandler'),
			entityInterceptor: getFieldValue(sessionFactoryBean, 'entityInterceptor'))

		newSessionFactoryFactory.afterPropertiesSet()

		def newSessionFactory = newSessionFactoryFactory.object

		['entityPersisters', 'collectionPersisters', 'identifierGenerators',
		 'namedQueries', 'namedSqlQueries', 'sqlResultSetMappings', 'imports',
		 'collectionRolesByEntityParticipant', 'classMetadata', 'collectionMetadata'].each { fieldName ->
			def field = ReflectionUtils.findField(ctx.sessionFactory.getClass(), fieldName)
			field.accessible = true
			field.set ctx.sessionFactory, new HashMap(field.get(newSessionFactory))
		}
	}

	private getFieldValue(sessionFactoryBean, String fieldName) {
		def field = ReflectionUtils.findField(sessionFactoryBean.getClass(), fieldName)
		field.accessible = true
		field.get sessionFactoryBean
	}
}
